"""Module for api keys."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from datetime import datetime


@dataclass
class ApiKey:
    """An api key contains the information of an api key"""

    name: str
    creation_date: datetime



"""Configure logging for Kukur."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0

import logging
import logging.handlers

from typing import Any, Dict


def configure(config: Dict[str, Any]):
    """Configure the Python logger."""
    log_format = "%(asctime)s %(levelname)s %(name)s %(threadName)s : %(message)s"
    if "logging" in config and "path" in config["logging"]:
        path: str = config["logging"]["path"]
        handler = logging.handlers.TimedRotatingFileHandler(
            path, when="D", backupCount=7
        )
        logging.basicConfig(level=logging.INFO, format=log_format, handlers=[handler])
    else:
        logging.basicConfig(level=logging.INFO, format=log_format)

    logging.getLogger("azure.identity").setLevel(logging.WARNING)
    logging.getLogger("azure.core").setLevel(logging.WARNING)



"""Connections to ADODB data sources from Kukur.

This requires an installation of pywin32 (LGPL).
"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

try:
    import adodbapi

    HAS_ADODB = True
except ImportError:
    HAS_ADODB = False

from kukur.exceptions import MissingModuleException
from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper
from kukur.source.sql import BaseSQLSource, SQLConfig


def from_config(
    data, metadata_value_mapper: MetadataValueMapper, quality_mapper: QualityMapper
):
    """Create a new ADODB data source from a configuration dict.

    Raises ADODBNotInstalledError when the adodbapi module is not available."""
    if not HAS_ADODB:
        raise MissingModuleException("pywin32", "adodb")

    config = SQLConfig.from_dict(data)

    return ADODBSource(config, metadata_value_mapper, quality_mapper)


class ADODBSource(BaseSQLSource):
    """An ADODB data source."""

    def __init__(
        self,
        config: SQLConfig,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):
        super().__init__(config, metadata_value_mapper, quality_mapper)
        if not HAS_ADODB:
            raise MissingModuleException("pywin32", "adodb")

    def connect(self):
        return adodbapi.connect(
            self._config.connection_string,
            {"timeout": self._config.query_timeout_seconds},
        )



"""Generic exceptions for Kukur data sources."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from typing import Optional


class KukurException(Exception):
    """Base class for all Exceptions thrown by Kukur."""


class InvalidDataError(KukurException):
    """Raised when the data itself is invalid."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"invalid data: {message}")


class InvalidSourceException(KukurException):
    """Raised when the source configuration is invalid."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"invalid source: {message}")


class MissingModuleException(KukurException):
    """Raised when a required Python module is not available."""

    def __init__(self, module_name: str, source_type: Optional[str] = None):
        if source_type is not None:
            KukurException.__init__(
                self,
                f'source type "{source_type}" requires Python package: "{module_name}"',
            )
        else:
            KukurException.__init__(
                self,
                f'missing Python package: "{module_name}"',
            )


class UnknownSourceException(KukurException):
    """Raised when the source is not known."""

    def __init__(self, source_name: str):
        KukurException.__init__(self, f"source does not exist: {source_name}")



"""Generic data loaders for file based columnar data."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import io

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Protocol, Union

try:
    from azure.identity import DefaultAzureCredential

    HAS_AZURE_IDENTITY = True
except ImportError:
    HAS_AZURE_IDENTITY = False

try:
    from azure.storage import blob

    HAS_AZURE_STORAGE_BLOB = True
except ImportError:
    HAS_AZURE_STORAGE_BLOB = False

from kukur.exceptions import InvalidDataError, KukurException, MissingModuleException


class UnknownLoaderError(KukurException):
    """Raised when the specified loader is unknown."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"unknown loader: {message}")


class Loader(Protocol):
    """Loader opens file-like objects containing data."""

    def open(self):
        """Open a file-like object from the loading source."""
        ...

    def has_child(self, subpath: str) -> bool:
        """Test if a file-like object exists at a subpath."""

    def open_child(self, subpath: str):
        """Open a file-like object at a subpath."""
        ...


class FileLoader:
    """Load data from files"""

    __path: Path
    __mode: str
    __files_as_path: bool

    def __init__(self, path: Path, mode: str, files_as_path: bool):
        self.__path = path
        self.__mode = mode
        self.__files_as_path = files_as_path

    def open(self):
        """Open the file at path for reading.

        If files_as_path is True, return the path."""
        if not self.__path.exists():
            raise InvalidDataError(f"'{self.__path}' does not exist")
        if self.__files_as_path:
            return self.__path
        return self.__path.open(mode=self.__mode)

    def has_child(self, name: str) -> bool:
        """Test if the file <name> is in the directory pointed to by <path>."""
        if not self.__path.is_dir():
            raise InvalidDataError(f'"{self.__path}" is not a directory')
        path = self.__path / name
        return path.exists()

    def open_child(self, name: str):
        """Open the file <name> that is in the directory path for reading.

        If files_as_path is True, return the path of the child."""
        if not self.__path.is_dir():
            raise InvalidDataError(f'"{self.__path}" is not a directory')
        path = self.__path / name
        if not path.exists():
            raise InvalidDataError(f"'{path}' does not exist")
        if self.__files_as_path:
            return path
        return path.open(mode=self.__mode)


@dataclass
class AzureBlobConfiguration:
    """Connection details for an Azure blob inside a container.

    When the connection string does not include authentication tokens, provide an identity."""

    connection_string: str
    container: str
    identity: Optional[str] = None


class AzureBlobLoader:
    """Load data from Azure blobs"""

    __mode: str
    __config: AzureBlobConfiguration
    __path: str

    def __init__(
        self,
        mode: str,
        config: AzureBlobConfiguration,
        path,
    ):
        self.__mode = mode
        self.__config = config
        self.__path = path
        if not HAS_AZURE_IDENTITY:
            raise MissingModuleException("azure-identity")
        if not HAS_AZURE_STORAGE_BLOB:
            raise MissingModuleException("azure-storage-blob")

    def open(self):
        """Read the contents of the Blob given by path in a BytesIO/StringIO buffer."""
        container_client = self._get_container_client()
        downloader = container_client.download_blob(self.__path)
        if "b" in self.__mode:
            buffer = io.BytesIO()
            buffer.write(downloader.content_as_bytes())
        else:
            buffer = io.StringIO()
            buffer.write(downloader.content_as_bytes().decode())
        buffer.seek(0)
        return buffer

    def has_child(self, name: str) -> bool:
        """Test if the given child blob exists."""
        container_client = self._get_container_client()
        blob_client = container_client.get_blob_client(self.__path + "/" + name)
        return blob_client.exists()

    def open_child(self, name: str):
        """Read the contents of the Blob given by path/name in a BytesIO buffer."""
        container_client = self._get_container_client()
        downloader = container_client.download_blob(self.__path + "/" + name)
        buffer: Union[io.BytesIO, io.StringIO]
        if "b" in self.__mode:
            buffer = io.BytesIO()
            buffer.write(downloader.content_as_bytes())
        else:
            buffer = io.StringIO()
            buffer.write(downloader.content_as_bytes().decode())
        buffer.seek(0)
        return buffer

    def _get_container_client(self):
        if self.__config.identity is not None and self.__config.identity == "default":
            credential = DefaultAzureCredential()
            client = blob.BlobServiceClient.from_connection_string(
                self.__config.connection_string, credential
            )
        else:
            client = blob.BlobServiceClient.from_connection_string(
                self.__config.connection_string
            )

        return client.get_container_client(self.__config.container)


def from_config(
    config: Dict[str, str], key="path", mode="rb", files_as_path=False
) -> Loader:
    """Create a loader from a configuration object.

    The main path to the data file that will be loaded is found by key in the config.
    The file will be opened with the given mode if supported by the loader.

    Some sources allow more efficient reading of files using memory mapping. To
    support this, set files_as_path to True, and a loader for local files will
    return the Paths instead of the opened files.
    """
    loader_type = config.get("loader", "file")
    if loader_type == "file":
        return FileLoader(Path(config[key]), mode, files_as_path)
    if loader_type == "azure-blob":
        if not HAS_AZURE_IDENTITY:
            raise MissingModuleException("azure-identity")
        if not HAS_AZURE_STORAGE_BLOB:
            raise MissingModuleException("azure-storage-blob")
        azure_config = AzureBlobConfiguration(
            config["azure_connection_string"],
            config["azure_container"],
            config.get("azure_identity"),
        )

        path = config[key]
        return AzureBlobLoader(mode, azure_config, path)

    raise UnknownLoaderError(loader_type)



"""The Kukur application dispatches data requests to the correct data source."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Union
from pathlib import Path

import json
import pyarrow as pa

from kukur import Metadata, SeriesSearch, SeriesSelector, SourceStructure
from kukur.exceptions import UnknownSourceException
from kukur.source import SourceFactory, SourceWrapper
from kukur.api_key.app import ApiKeys

from kukur.repository import MigrationRunner, RepositoryRegistry


class Kukur:
    """Kukur queries the sources it's configured with for time series data and metadata.

    It implements the Source interface."""

    __repository: RepositoryRegistry

    def __init__(self, config: Dict[str, Any]):
        self.__source_factory = SourceFactory(config)
        self.__repository = RepositoryRegistry(
            data_dir=Path(config.get("data_dir", "."))
        )
        migration_runner = MigrationRunner()
        migration_runner.register(self.__repository.api_key().migrations())
        migration_runner.migrate()

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[SeriesSelector, Metadata], None, None]:
        """Return all time series or even the metadata of them in this source matching the selector."""
        return self._get_source(selector.source).search(selector)

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Return metadata for the given time series."""
        return self._get_source(selector.source).get_metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""
        return self._get_source(selector.source).get_data(
            selector, start_date, end_date
        )

    def get_plot_data(
        self,
        selector: SeriesSelector,
        start_date: datetime,
        end_date: datetime,
        interval_count: int,
    ) -> pa.Table:
        """Return plot data for the given time series in the given time period.

        Normal data is returned for sources that do not support plot data."""
        return self._get_source(selector.source).get_plot_data(
            selector, start_date, end_date, interval_count
        )

    def get_source_structure(
        self, selector: SeriesSelector
    ) -> Optional[SourceStructure]:
        """Return the structure of a source."""
        return self._get_source(selector.source).get_source_structure(selector)

    def get_api_keys(self) -> ApiKeys:
        """Return the api keys."""
        return ApiKeys(self.__repository)

    def list_sources(self, *_) -> List[bytes]:
        """Return all the configured sources."""
        sources = self.__source_factory.get_source_names()
        return [json.dumps(sources).encode()]

    def _get_source(self, source_name: str) -> SourceWrapper:
        source = self.__source_factory.get_source(source_name)
        if source is None:
            raise UnknownSourceException(source_name)
        return source



"""Kukur source for PI Data Archives using PI Web API."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from .piwebapi_da import from_config

__all__ = ["from_config"]



"""Read the Kukur configuration."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import glob
import logging

import toml

from kukur.exceptions import KukurException


logger = logging.getLogger(__name__)


class InvalidIncludeException(KukurException):
    """Raised when the include configuration is invalid."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"invalid include: {message}")


def from_toml(path):
    """Read the configuration from a TOML file, processing includes."""
    config = _read_toml(path)
    for include_options in config.get("include", []):
        if "glob" not in include_options:
            raise InvalidIncludeException('"glob" is required')
        for include_path in glob.glob(include_options["glob"]):
            include_config = _read_toml(include_path)
            for k, v in include_config.items():
                if k not in config:
                    config[k] = v
                elif isinstance(config[k], list):
                    config[k].extend(v)
                elif isinstance(config[k], dict):
                    config[k].update(v)
                else:
                    config[k] = v
    return config


def _read_toml(path):
    try:
        return toml.load(path)
    except Exception as err:
        logger.error("error in %s", path)
        raise err



"""Package Kukur for publishing."""

import os
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kukur",
    version=os.environ.get('KUKUR_VERSION', '0.0.0'),
    author="Timeseer.AI",
    author_email="pypi@timeseer.ai",
    description="Kukur makes time series data and metadata available to the Apache Arrow ecosystem.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    url="https://kukur.timeseer.ai/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=setuptools.find_packages(exclude="tests"),
    install_requires=[
          'pyarrow',
          'python-dateutil',
          'pytz',
          'toml',
      ],
    extras_require={
        "adodb": ["pywin32"],
        "cratedb": ["crate"],
        "delta": ["deltalake"],
        "influxdb": ["influxdb"],
        "odbc": ["pyodbc"],
        "piwebapi": ["requests", "requests-kerberos"],
    },
    python_requires='>=3.8',
    package_data={
        'kukur': ['py.typed'],
    },
    entry_points={
        "console_scripts": ["kukur = kukur.cli:_run"],
    },
)
